#include "Rays.h"

Rays::Rays() { }

/**
 * Initializes the rays and polygons with the correct shaders.
 */
void Rays::init(const ObjectCollection *testableObjects, shader_prog *in_lineShader, shader_prog *in_polgyonShader) {
    lineShader = in_lineShader;
    polygonShader = in_polgyonShader;

    lineShader->activate();
    initRays(testableObjects);
    printf("Rays: %i \n", rays.size());

    polygonShader->activate();
    initShadowPolygon();
}

/**
 * This function should populate the rays vector with Ray instances.
 * We also create the VAO and VBO for drawing the rays.
 */
void Rays::initRays(const ObjectCollection *testableObjects) {
    GLuint vaoHandle;
    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);
    std::vector<glm::vec3> rayVertices = std::vector<glm::vec3>();

    /**
     * --Task--
     * Loop through the testableObejcts. For each object:
     * - Loop through its world vertices. For each vertex:
     * -- Create 3 rays with shifts -0.001, 0.0, 0.001 radians.
     * -- Push the created ray wo the rays vector.
     * -- Put something in the rayVertices in order to send that as attributes.
     */
    for (auto& object : *testableObjects) {

        //...

    }

    GLuint vboHandle = lineShader->attributeVectorVec3("position", rayVertices);

    raysVBO = vboHandle;
    raysVAO = vaoHandle;
}

/**
 * Updates the rays: trace them against the testableObjects.
 * Updates the raysVBO with the new data.
 * Calls the update on the shadow polygon.
 */
void Rays::update(const ObjectCollection* testableObjects) {
    std::vector<glm::vec3> rayVertices = std::vector<glm::vec3>(); //This will hold our new data

    for (auto &ray : rays) {

        /**
         * --Task--
         * Update the directions (ray.dir) for rays.
         * Think about what the updated direction vector is:
         * 	- From where to where
         * 	- In which space are those coordinates (object's model matrix is "matrix" or "matrixWorld")
         *  - Do not forget to normalize it (Three.js wants it to be normalized for the trace)
         *
         * Then update the hit of the ray (ray.hit) with the result from the rayTrace against testableObjects.
         *
         * Push the start and hit into rayVertices for an update in the VBO.
         */
    }

    //Here we send the rayVertices as the new data for the VBO.
    lineShader->activate();
    glBindVertexArray(raysVAO);
    glBindBuffer(GL_ARRAY_BUFFER, raysVBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat)*rayVertices.size()*3, &rayVertices[0], GL_DYNAMIC_DRAW);

    updateShadowPolygon();

    return;
}

/**
 * Draws the rays with the line shader.
 * Then calls the drawShadowPolygon().
 */
void Rays::draw() {

    lineShader->activate();
    lineShader->uniformMatrix4fv("modelMatrix", glm::translate(glm::mat4(), glm::vec3(0.0f, 0.0f, 2.0f)));
    lineShader->uniform3f("color", 0.0, 1.0, 0.0);
    glBindVertexArray(raysVAO);
    glDrawArrays(GL_LINES, 0, rays.size() * 2);
    glBindVertexArray(0);

    drawShadowPolygon();
}

/**
 * When the mouse moves, we update the start for each ray.
 */
void Rays::mouseMove(double x, double y, double width, double height) {
    for (Ray &ray : rays) {
        ray.start = glm::vec3(x - width / 2, -y + height / 2, 0.0);
    }
}

/**
 * Ray tracing for a ray against the testableObjects.
 */
glm::vec3 Rays::rayTrace(Ray &ray, const ObjectCollection* testableObjects) {
    float dist = INFINITY;
    float currentDist;

    /**
     * --Task--
     * Loop over the testableObjects. For each object:
     *  - Loop over its vertices. For each vertex:
     *  -- Call rayToLineSegment with the ray and correct vertices (Hint: assign to currentDist).
     *  -- Find the closest intersection of the ray and line segments.
     *  -- Note: Be sure to account for all line segments that the vertices form.
     */


    return ray.start + dist * ray.dir; //We return the hit point
}

/**
 * This tests the ray and line segment intersection.
 * This algorithm is based on the Mller-Trumbore algorithm we used in the Ray Trace task.
 * Other existing algorithms I found turned out to be unstable for some reason.
 * So I decided to use the previous approach with rays and line segments in 2D.
 */
float Rays::rayToLineSegment(const Ray &ray, const glm::vec3 &lineStart, const glm::vec3 &lineEnd){
    glm::vec2 rayDir = glm::vec2(ray.dir);
    glm::vec2 segDir = glm::vec2(lineEnd) - glm::vec2(lineStart);

    float eps = 0.0001f;
    glm::vec2 p = glm::vec2(lineStart - ray.start);

    float det = rayDir.x * (-segDir.y) - rayDir.y * (-segDir.x);
    float detSeg = rayDir.x * p.y - rayDir.y * p.x;

    float u = detSeg / det;
    if (u < -eps || u > 1.0 + eps) {

        return INFINITY;
    }

    float detRay = p.x * (-segDir.y) - p.y * (-segDir.x);
    float t = detRay / det;

    return (t >= 0.0) ? t : INFINITY;
}

/**
 * Initializes the shadowPolygon with some vertices.
 */
void Rays::initShadowPolygon() {
    GLuint vaoHandle;

    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);

	std::vector<glm::vec3> vertices = std::vector<glm::vec3>();

	/**
	 * --Task--
	 * Push some vertices into the vertices vector.
	 * They can be empty and / or ray hit positions, we will update them later.
	 */



    glm::mat4 model = glm::mat4();

    shadowPolygonVBO = polygonShader->attributeVectorVec3("position", vertices);
    shadowPolygon = new Object3D(vertices, model, vaoHandle);

    return;
}

/**
 * Updates the shadowPolygon with the new vertices.
 */
void Rays::updateShadowPolygon() {
    /**
     * --Task--
     * Call std::sort() on the rays vector. Be sure to implement the comparison operator for the rays first.
     */


    /**
     * --Task--
     * Set the vertices correctly for the shadowPolygon (use setVertex(glm::vec3 vertex, int index)).
     * Remember how we construct the shadow polygon for the correct assignment of vertices.
     */


    //After setting the new coordinates, we update them
    shadowPolygon->updateWorldVertices();
}

/**
 * Draws the shadowPolygon with its world space vertices.
 */
void Rays::drawShadowPolygon() {
    polygonShader->activate();
    polygonShader->uniformMatrix4fv("modelMatrix", glm::mat4());
    glBindVertexArray(shadowPolygon->getVAO());
    glBindBuffer(GL_ARRAY_BUFFER, shadowPolygonVBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat)*shadowPolygon->getVertexCount()*3, &(shadowPolygon->getWorldVertices())[0], GL_STATIC_DRAW);
    glDrawArrays(GL_TRIANGLE_FAN, 0, shadowPolygon->getVertexCount());
}
