#include "Scene.h"
#include <iostream>
#include <stdexcept>

Scene::Scene() {}

/**
 * Constructs the scene objects. Assigns the line and floor shaders.
 * Floor needs to know the width (= height) of our viewport.
 */
void Scene::init(shader_prog* in_lineShader, shader_prog* in_floorShader, float width) {
    lineShader = in_lineShader;
    floorShader = in_floorShader;

    lineShader->activate(); //Box, letters and triangels will use the line shader
    createBox(width);
    createLetterC();
    createLetterG();
    createTriangles();

    floorShader->activate(); //Foor uses another shader to render with a texture
    createFloor(width);

    try {
        std::random_device rd;
        generator = std::mt19937(rd());
        distribution = std::uniform_real_distribution<>(1.5, 2.5);
    } catch( std::exception &err ) { //Had some problems on an older CodeBlocks version
        std::cerr << "Error initializing RNG: " << err.what() << std::endl;
    }

}

/**
 * Updating a scene means updating every object the scene knows about.
 */
void Scene::update(double dt) {
    for (auto& object : testableObjects) {
        object->update(dt);
    }
}

/**
 * General function to draw a Object3D with the line shader.
 */
void Scene::drawLineObject(Object3D& object) {
    lineShader->activate();
    lineShader->uniformMatrix4fv("modelMatrix", glm::translate(object.getModel(), glm::vec3(0.0f, 0.0f, 2.0f)));
    lineShader->uniform3f("color", 1.0, 0.0, 0.0);
    glBindVertexArray(object.getVAO());
    glDrawArrays(GL_LINE_LOOP, 0, object.getVertexCount());
}

/**
 * Drawing the scene.
 */
void Scene::draw() {
    //Remove this to see object outlines and rays only
    /**
     * --Task--
     * Enable the blending and use multiplicative blending to draw the floor.
     * The shadow polygon has been drawn, need to blend the floor with it.
     * After drawing the floor, disable the blending.
     */

    //drawFloor();



    //We draw each line object
    for (auto& object : testableObjects) {
        drawLineObject(*object);
    }
}

/**
 * This draw the floor with the floor shader.
 */
void Scene::drawFloor() {
    floorShader->activate();
    floorShader->uniformMatrix4fv("modelMatrix", glm::translate(glm::mat4(), glm::vec3(0.0f, 0.0f, 2.0f)));
    glBindVertexArray(floorVAO);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, floorTextureHandle);
    glDrawArrays(GL_TRIANGLES, 0, 6);
}

/**
 * Creates a box with the given width (= height) that encapsulates the scene.
 */
Object3D Scene::createBox(float width) {
    GLuint vaoHandle;

    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);

	std::vector<glm::vec3> vertices = std::vector<glm::vec3>();
    vertices.push_back(glm::vec3(  width / 2      ,  width / 2 - 0.5, 0 ));
    vertices.push_back(glm::vec3( -width / 2 + 0.5,  width / 2 - 0.5, 0 ));
    vertices.push_back(glm::vec3( -width / 2 + 0.5, -width / 2      , 0 ));
    vertices.push_back(glm::vec3(  width / 2      , -width / 2      , 0 ));

    glm::mat4 model = glm::mat4();
    lineShader->attributeVectorVec3("position", vertices);

    //Box is added to the testableObjects collection.
    testableObjects.emplace_back(new Object3D(vertices, model, vaoHandle));

    return *testableObjects.back();
}

/**
 * Creates a floor that encapsulates the scene.
 */
GLuint Scene::createFloor(float width) {
    GLuint vaoHandle;

    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);

	std::vector<glm::vec3> vertices = std::vector<glm::vec3>();
    vertices.push_back(glm::vec3(  width / 2      ,  width / 2 - 0.5, 1.0 ));
    vertices.push_back(glm::vec3( -width / 2 + 0.5,  width / 2 - 0.5, 1.0 ));
    vertices.push_back(glm::vec3( -width / 2 + 0.5, -width / 2      , 1.0 ));

    vertices.push_back(glm::vec3( -width / 2 + 0.5, -width / 2      , 1.0 ));
    vertices.push_back(glm::vec3(  width / 2      , -width / 2      , 1.0 ));
    vertices.push_back(glm::vec3(  width / 2      ,  width / 2 - 0.5, 1.0 ));

    floorShader->attributeVectorVec3("position", vertices);

    std::vector<glm::vec2> uvs = std::vector<glm::vec2>();
    uvs.push_back(glm::vec2(1.0, 1.0));
    uvs.push_back(glm::vec2(1.0, 0.0));
    uvs.push_back(glm::vec2(0.0, 0.0));

    uvs.push_back(glm::vec2(0.0, 0.0));
    uvs.push_back(glm::vec2(0.0, 1.0));
    uvs.push_back(glm::vec2(1.0, 1.0));
    floorShader->attributeVectorVec2("uv", uvs);

    floorTextureHandle = load_texture(GL_TEXTURE_2D, GL_SRGB, "data/floor2.jpg");
    //floorTextureHandle = load_texture(GL_TEXTURE_2D, GL_SRGB, "data/pong.png");

    floorVAO = vaoHandle;

    return floorVAO;
}

/**
 * Creates the letter C.
 */
Object3D Scene::createLetterC() {
    GLuint vaoHandle;

    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);

	std::vector<glm::vec3> vertices = std::vector<glm::vec3>();
    vertices.push_back(glm::vec3(  10.0f,  10.0f, 0.0f ));
    vertices.push_back(glm::vec3(   5.0f,  15.0f, 0.0f ));
    vertices.push_back(glm::vec3( -10.0f,  15.0f, 0.0f ));
    vertices.push_back(glm::vec3( -20.0f,   0.0f, 0.0f ));
    vertices.push_back(glm::vec3( -10.0f, -15.0f, 0.0f ));
    vertices.push_back(glm::vec3(   5.0f, -15.0f, 0.0f ));
    vertices.push_back(glm::vec3(  10.0f, -10.0f, 0.0f ));
    vertices.push_back(glm::vec3(   5.0f, -12.0f, 0.0f ));
    vertices.push_back(glm::vec3(  -5.0f, -10.0f, 0.0f ));
    vertices.push_back(glm::vec3( -12.0f,   0.0f, 0.0f ));
    vertices.push_back(glm::vec3(  -5.0f,  10.0f, 0.0f ));
    vertices.push_back(glm::vec3(   5.0f,  12.0f, 0.0f ));

    glm::mat4 model = glm::mat4();
    model = glm::translate(model, glm::vec3(-70.0f, 0.0f, 0.0f));
    model = glm::scale(model, glm::vec3(5.0f, 5.0f, 1.0f));

    lineShader->attributeVectorVec3("position", vertices);

    //This is added to the testableObjects collection.
    testableObjects.emplace_back(new Object3D(vertices, model, vaoHandle));

    return *testableObjects.back();
}

/**
 * Creates the letter G.
 */
Object3D Scene::createLetterG() {
    GLuint vaoHandle;

    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);

	std::vector<glm::vec3> vertices = std::vector<glm::vec3>();
    vertices.push_back(glm::vec3(  10,  10, 0 ));
    vertices.push_back(glm::vec3(   5,  15, 0 ));
    vertices.push_back(glm::vec3( -10,  15, 0 ));
    vertices.push_back(glm::vec3( -20,   0, 0 ));
    vertices.push_back(glm::vec3( -10, -15, 0 ));
    vertices.push_back(glm::vec3(   5, -15, 0 ));
    vertices.push_back(glm::vec3(  10, -10, 0 ));
    vertices.push_back(glm::vec3(  10,  -5, 0 ));
    vertices.push_back(glm::vec3(   5,   2, 0 ));
    vertices.push_back(glm::vec3(  -5,  -3, 0 ));
    vertices.push_back(glm::vec3(   3,  -2, 0 ));
    vertices.push_back(glm::vec3(   6,  -7, 0 ));
    vertices.push_back(glm::vec3(  -5, -10, 0 ));
    vertices.push_back(glm::vec3( -12,   0, 0 ));
    vertices.push_back(glm::vec3(  -5,  10, 0 ));
    vertices.push_back(glm::vec3(   5,  12, 0 ));

    glm::mat4 model = glm::mat4();
    model = glm::translate(model, glm::vec3(100.0, 0.0, 0.0));
    model = glm::scale(model, glm::vec3(5.0, 5.0, 1.0));

    lineShader->attributeVectorVec3("position", vertices);

    //This is added to the testableObjects collection.
    testableObjects.emplace_back(new Object3D(vertices, model, vaoHandle));

    return *testableObjects.back();
}

/**
 * Creates a number of triangles at some positions.
 */
void Scene::createTriangles() {
    std::vector<glm::vec2> trianglePositions = std::vector<glm::vec2>();
    trianglePositions.push_back(glm::vec2( -120,  120));
    trianglePositions.push_back(glm::vec2(  -10,  170));
    trianglePositions.push_back(glm::vec2(   30,   90));
    trianglePositions.push_back(glm::vec2(  140,  110));
    trianglePositions.push_back(glm::vec2( -140, -140));
    trianglePositions.push_back(glm::vec2(  -50, -170));
    trianglePositions.push_back(glm::vec2(   60, -120));
    trianglePositions.push_back(glm::vec2(  140, -110));

    for (auto trianglePosition : trianglePositions) {
        //Each triangle is added to the testableObjects collection.
        testableObjects.emplace_back(createTriangle(trianglePosition));
    }
}

/**
 * Creates one triangle.
 * Assigns the given position to that triangle's model transformation.
 */
RotatingObject3D* Scene::createTriangle(glm::vec2 position) {
    GLuint vaoHandle;

    glGenVertexArrays(1, &vaoHandle);
    glBindVertexArray(vaoHandle);

    //Hopefully you did something like this in the very first week.
    float sqrt3 = glm::sqrt(3.0);
    float r = 5.0; //Incircle radius
	std::vector<glm::vec3> vertices = std::vector<glm::vec3>();
    vertices.push_back(glm::vec3(          0, -2.0 * r, 0 )); //Top vertex
    vertices.push_back(glm::vec3( -sqrt3 * r,        r, 0 )); //Left vertex
    vertices.push_back(glm::vec3(  sqrt3 * r,        r, 0 )); //Right vertex

    glm::mat4 model = glm::mat4();
    model = glm::translate(model, glm::vec3(position.x, position.y, 0.0f));
    model = glm::scale(model, glm::vec3(2.0, 2.0, 1.0));

    lineShader->attributeVectorVec3("position", vertices);

    //Triangles are instances of RotatingObject3D. They have some rotation speed.
    RotatingObject3D* object = new RotatingObject3D(vertices, model, vaoHandle);
    float speed = (distribution(generator) + 0.5) * (distribution(generator) > ((distribution.b() - distribution.a()) / 2.0) ? -1.0f : 1.0f);
    object->setRotationSpeed(speed);

    return object;
}

/**
 * Getter for the testableObjects collection.
 */
ObjectCollection* Scene::getTestableObjects() {

    return &testableObjects;
}
