// ---------------------------- Includes -------------------------- //
#include <stdlib.h>         // C++ standard library
#include <stdexcept>
#include <stack>            // We use the standard C++ stack implementation to create model matrix stacks
#include <unistd.h>         // Threading (for sleep)
#include <stdio.h>          // Input/Output
#include <memory>
#include <GLEW/glew.h>      // OpenGL Extension Wrangler -
#include <GLFW/glfw3.h>     // Windows and input
#include <glm/glm.hpp>      // OpenGL math library
#include <glm/gtc/constants.hpp>
#include <glm/gtx/intersect.hpp>
#include <typeinfo>

#include "shader_util.h"    // Utility methods to keep this file a bit shorter.
#include "texture_util.h"   // Utility for texture loading and binding

#include "Object3D.h"       // 3D objects
#include "RotatingObject3D.h" // Rotating 3D objects
#include "Rays.h"           // Rays, intersection testing, shadow polygon
#include "Scene.h"          // Scene

// --- Load the shaders declared in glsl files in the project folder ---//
shader_prog floorShader("shaders/floor.vert.glsl", "shaders/floor.frag.glsl");
shader_prog lineShader("shaders/line.vert.glsl", "shaders/line.frag.glsl");
shader_prog polygonShader("shaders/polygon.vert.glsl", "shaders/polygon.frag.glsl");

float width = 400.0f;
float height = 400.0f;
Rays rays = Rays();
Scene scene = Scene();

// ---------------------------- Input -------------------------- //
// This method is called when keyboard event happens.
// Sets GLFW window should close flag to true, when escape key is pressed.
static void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
    if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) {
        glfwSetWindowShouldClose(window, GL_TRUE);
    }
}
static void mouse_move_callback(GLFWwindow* window, double x, double y) {
    rays.mouseMove(x, y, width, height);
}

// ---------------------------- Main -------------------------- //
int main(int argc, char *argv[]) {
    GLFWwindow *win;
    if (!glfwInit()) {
        exit (EXIT_FAILURE);
    }

    win = glfwCreateWindow(width, height, "2D Shadows", NULL, NULL);

    if (!win) {
        glfwTerminate();
        exit(EXIT_FAILURE);
    }

    glfwMakeContextCurrent(win);

    //Some extensions are marked as experimental. To use the latest
    //version of OpenGL supported, we have to set this flag to true.
    glewExperimental = GL_TRUE;
    GLenum status = glewInit();

    if(status != GLEW_OK) {
        fprintf(stderr, "Error: %s\n", glewGetErrorString(status));
    }

    const GLubyte* renderer = glGetString (GL_RENDERER); // get renderer string
    const GLubyte* version = glGetString (GL_VERSION); // version as a string
    printf ("Renderer: %s\n", renderer);
    printf ("OpenGL version supported %s\n", version);

    glfwSetKeyCallback(win, key_callback);
    glfwSetCursorPosCallback(win, mouse_move_callback);

    lineShader.use();   //Compile the shaders
    polygonShader.use();
    floorShader.use();

    glm::mat4 projection = glm::ortho(-width / 2.0f, width / 2.0f, -height / 2.0f, height / 2.0f, 0.1f, 100.f);
    glm::mat4 view = glm::lookAt(
        glm::vec3(0.0, 0.0, 15.0), //Position
        glm::vec3(0.0, 0.0, 0.0),  //LookAt
        glm::vec3(0.0, 1.0, 0.0)   //Up
    );

    //Initialize the scene and rays
    scene.init(&lineShader, &floorShader, width);
    rays.init(scene.getTestableObjects(), &lineShader, &polygonShader);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

    float deltaTime, currentFrameTime;
    float lastFrameTime = 0.0f;

    while (!glfwWindowShouldClose(win)) {
        currentFrameTime = glfwGetTime();
        deltaTime = currentFrameTime - lastFrameTime;

        scene.update(deltaTime);                // Update the scene (rotate triangles)
        rays.update(scene.getTestableObjects());// Update the rays, find new intersections, update ray lines and polygon

        // All the shaders get the projection and view matrices
        lineShader.activate();
        lineShader.uniformMatrix4fv("projectionMatrix", projection);
        lineShader.uniformMatrix4fv("viewMatrix", view);

        polygonShader.activate();
        polygonShader.uniformMatrix4fv("projectionMatrix", projection);
        polygonShader.uniformMatrix4fv("viewMatrix", view);

        floorShader.activate();
        floorShader.uniformMatrix4fv("projectionMatrix", projection);
        floorShader.uniformMatrix4fv("viewMatrix", view);

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        rays.draw();    //Draw rays (incl. shadow polygon) first, because we will blend the floor with the shadow polygon
        scene.draw();   //Draw the scene objects

        glfwSwapBuffers(win);
        usleep(10000);
        glfwPollEvents();

        lastFrameTime = currentFrameTime;
    }

    glfwTerminate();
    exit(EXIT_SUCCESS);

    return 0;
}
